@import url('https://fonts.googleapis.com/css2?family=Roboto:ital,wght@0,100;0,300;0,400;0,500;0,700;0,900;1,100;1,300;1,400;1,500;1,700;1,900&display=swap');

:root {
    --bulma-text: black;
}

html {
    font-size: 14px;
    font-family: 'Roboto', sans-serif;
    background-color: white;
}

.dropdown {
    width: 100%;;
}

#navbar {
    height: 50px;
    font-weight: bold;
}

#upload-data {
    font-family: monospace;
    cursor: pointer;
}

#metadata-viewer {
    font-size: 14px;
    max-height: 300px;
    overflow-y: scroll;
    font-family: monospace;
}

.button {
    margin: 5px;
}